/*
 *	Copyright (C) 2008, 2009. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvdesignergui.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef DESIGNERGUI_H
#define DESIGNERGUI_H

#include <QObject>
#include <QList>
#include <QString>
#include <QMap>
class QXmlStreamReader;
#include "../qvgui.h"
#include "facade/itemproperties.h"
#include "facade/itemfactory.h"
#include "slate/slatewindow.h"
#include "slate/informer.h"
#include <QMetaType>
class QVPropertyContainer;
class QVPropertyContainerChange;


/*!
@class QVDesignerGUI qvblockprogramming/qvguiblocks/qvdesigner/qvdesignergui.h QVDesignerGUI
@brief Advanced graphic block oriented programming interface widget for QVision applications.

Class @ref QVDesignerGUI is used in QVision applications to create an advanced GUI that offers the user the possibility to control the execution of the different workers registered in the application, the flow of the input video camera objects, modify the input parameters defined in them (and which are not linked to the output of other workers, in whose case they obviously should not be controlled directly by the user), and modify the application's structure with a graphic oriented programming widget in run time.

A single instance object of the QVDesignerGUI (or others QVGUI classes, like QVDefaultGUI) class should be created right after the @ref QVApplication object in the <i>main</i> function, and before the call to <i>exec()</i>:

\code
int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		"Example program for QVision library. Obtains several features from input video frames."
		);
	
	QVDesignerGUI interface;

	[...]

	return app.exec();
	}
\endcode

The model control interface of QVDesignerGUI is the same of QVDefaultGUI interface. Its application get the same execution's parameter, and it uses the same input and output parameters's widgets (but QVDesigner places them differently).

The following screenshot shows the main QVDesignerGUI for a QVision example application:

\image html designer-slate.png

<!-- Widget areas in the QVDefaultGUI -->
There are four main widget areas in the @ref QVDefaultGUI window:

<!-- --- -->
@section MenuArea A. Menu area

Contains the File, Edit and Insert menus.

In the <i>File</i> menu we can find the following actions:

<table>
<tr><td><b>Export</b></td> <td> save the slate's model as a .cpp file. These file contains the default QVision code for the model.</td></tr>
<tr><td><b>Save As</b></td> <td> save the slate's model as a .xml file. These file contanis a XML representation of the model, that can be loaded by QVDesigner.</td></tr>
<tr><td><b>Load</b></td> <td>load a .xml file that contains a XML representation of a model, in the QVDesigner format.</td></tr>
<tr><td><b>Exit</b></td> <td>Exit the application.</td></tr>
</table>

In the <i>Edit</i> menu we can find the following actions:

<table>
<tr><td><b>Asinchronous Link Mode</b></td> <td>sets the QVision link type as asinchronous. The following links done in the model will be asinchronous links.</td></tr>
<tr><td><b>Sinchronous Link Mode</b></td> <td>sets the QVision link type as sinchronows. The following links done in the model will be asinchronous links.</td></tr>
<tr><td><b>Sequential Link Mode</b></td> <td>sets the QVision link type as sequential. The following links done in the model will be asinchronous links.</td></tr>
<tr><td><b>Join</b></td> <td>join the selected blocks into a new group block. the group block shows its internal block and adapts theirs externals links.</td></tr>
<tr><td><b>Delete</b></td> <td>delete the selected items (blocks, groups and/or links).</td></tr>
<tr><td><b>Bring to Front</b></td> <td>bring the selected item representation front of the rest of items.</td></tr>
<tr><td><b>Send to Back</b></td> <td>send the selected item representation back of the rest of items.</td></tr>
<tr><td><b>Properties...</b></td> <td>show the selected item properties: properties values, properties controls and item controls.</td></tr>
</table>

In the <i>Insert</i> menu we can find all items types that we can insert in the slate, by clicking it. We can find them organized in the following groups:

<table>
<tr><td><b>IPP: Filtering Functions</b></td> <td> contains items that executes IPP filtering functions, like Lowpass, Gauss, Laplace, Median, Min and Max filtering.</td></tr>
<tr><td><b>IPP: Image Arithmetic and Logical Operations</b></td> <td> contains items that executes aritmetic and logic functions over images.</td></tr>
<tr><td><b>IPP: Image Color Conversion</b></td> <td> contains items that executes IPP image color conversion functions.</td></tr>
<tr><td><b>IPP: Image Data Exchange and Initialization Functions</b></td> <td> contains items that executes IPP image data exchange and initialization funtions.</td></tr>
<tr><td><b>IPP: Image Estatistics Functions</b></td> <td> contains items that executes IPP stadistic functions.</td></tr>
<tr><td><b>IPP: Threshold and Compare Operations</b></td> <td> contains items that executes IPP threshold and compare funtions.</td></tr>
<tr><td><b>Input node</b></td> <td> contains QVision input items, like cameras and point generators.</td></tr>
<tr><td><b>Output node</b></td> <td> contains QVision output items, like canvas, plots and recorders.</td></tr>
<tr><td><b>QVision workers</b></td> <td> contains QVision advanced processing items.</td></tr>
<tr><td><b>User node</b></td> <td> contains the user items, defined in the user's application.</td></tr>
</table>

<!-- --- -->
@section ActionBar B. Action bar

The action bar contains buttons for the more frecuently actions: Asinchronous Link Mode, Sinchronous Link Mode, Sequential Link Mode, Join, Delete, Bring to Front, Send to Back. The action bar includes two buttons for running and stopping the sistem too:

<table>
<tr><td><b>Run</b> </td> <td> @image html qvdesignergui_run.png </td>
<td> Runs the processing sistem. When we open an application it start running.</td></tr>

<tr><td><b>Stop</b> </td> <td> @image html qvdesignergui_stop.png </td>
<td>Stops the processing sistem. In order to make designer changes, the sistem must be stopped.</td></tr>
</table>

<!-- --- -->
@section Slate C. Slate

We can do a lot of operations from the slate by mouse clicking or mouse dragging.

<b>The slate back</b>

By right clicking in the slate back, we can open an action menu whith the allowed operations. By mouse dragging over the slate back, we can select several items.

<b>Items</b>

The items are divided in three parts: the head the body and the link points (placed in the left and the right of the body).

We can select an item by left clicking on it, and we can open an allowed actions menu by right clicking the item. By double clicking the item's body we open a properties window. By double clicking the item's head we open a remane window

The items can be moved from the slate by mouse dragging. But if we are dragging over a lincking point, we are creating a new link, the link will be created if we drag the link over other allewed lincking point and we drop it.

<b>Groups</b>

The groups and the items are similar. They have a head (that is out), a body and linking points. We can do the same actions less what open the properties window (groups have not it). In a group's body we can see its internal items, that have been reduced. We can operate with the internal items normally. The scroll wheel allows us to do zoom over the slate. There is an arrow in the group's body right-bottom side, this arrow lets us to change he group view (by double clicking it), the group view can change from internal items view to abstract view. Abstract group's view hide its internal items and only show the name of the properties lincken outside the group.

<!-- --- -->
@section StateBar C. State bar

The state bar is under the slate, abd show error and sistem messages.

@ingroup qvgui
*/
class QVDesignerGUI: QObject, QVGUI
{ // Clase que se encargará de hacer de intermediario entre la vista y el modelo (el slate y la QVAplication)
	Q_OBJECT
	
	public:

		/// @brief Constructs a QVDesignerGUI object.
		// /// @param parent (Optional) parent for the created widget
		QVDesignerGUI();

#ifndef DOXYGEN_IGNORE_THIS
		//////////////////////////////////////////// funcionalidad de gui de la qvision ////////////////////////////////////////////

		/// @brief initialice the QVGUI object.
		virtual void init();
	
		/// @brief show the QVGUI object.
		void show();

		//////////////////////////////////////////// obtención de los items disponibles ////////////////////////////////////////////

		// devuelve la lista de Items disponibles en el sistema, cada uno de estos items es un elemento del diseño de los programas de QVision
		QMap<QString, QList<QString> > getItemTypes() const;
	
		// devuelve la lista de Items de entrada disponibles en el sistema, cada uno de estos items es un elemento del diseño de los programas de QVision
		QMap<QString, QList<QString> > getInputItemTypes() const;
	
		// devuelve la lista de Items intermedios disponibles en el sistema, cada uno de estos items es un elemento del diseño de los programas de QVision
		QMap<QString, QList<QString> > getMiddleItemTypes() const;
	
		// devuelve la lista de Items de salida disponibles en el sistema, cada uno de estos items es un elemento del diseño de los programas de QVision
		QMap<QString, QList<QString> > getOutputItemTypes() const;

		QString getCppText() const;

		QString getXMLText();

		bool loadXML(QString text);


		////////////////////////////////////// modificación del entorno común (slate + qvision) //////////////////////////////////////

		/// @todo Revisar, función para añadir. Esta función permite registrar nuevos tipos de blocks en el ámbito de nombres de la aplicación, para que
		/// el designer pueda crear clases de ese tipo.
		// crea y registra un item, del tipo especificado y nombre concreto, en el sistema. Devuelve el id del item (0 si no lo ha creado)
		template<typename T> static uint registerUserType(const QString name)
			{ return qRegisterMetaType<T>(name.toAscii().data()); }

		// crea y registra un item, del tipo especificado y nombre concreto, en el sistema. Devuelve el id del item (0 si no lo ha creado)
		uint addItem(const QString type, const QString name, uint lastId = 0);
	
		// crea y registra un link entre dos items del sistema
		bool addLink(const uint fromId, const QString fromProp, const uint toId, const QString toProp, const bool synchronous, const bool sequential);

		// borra y desregistra un item
		bool delItem(const uint id);
	
		// borra y desregistra un link
		bool delLink(const uint fromId, const QString fromProp, const uint toId, const QString toProp);

		// establece el valor de la propiedad de un item del sistema
		template <class Type> bool setProperty(const uint fromId, const QString fromProp, const Type &value);

		bool setName(const uint id, const QString name);

		void showProperties(const uint id);

		///////////////////////////////////////////////////// control del entorno /////////////////////////////////////////////////////

		// lanza un sistema como el compuesto en la pizarra
		void run();
	
		// para el sistema lanzado
		void stop();

	public slots:
		void quit();
		void processChange(QVPropertyContainerChange change);
		void dialogChange(QVPropertyContainerChange change);

	private:
		bool createDialog(const uint itemId);
		void deleteDialog(const uint itemId);
		QString getLinkName(QVPropertyContainerChange change);
		QString getLinkName(QString fromName, QString fromProp, QString toName, QString toProp);
		QString getAbsPropName(QString fromName, QString fromProp);
		void createGroups(QList<GroupInfo> &lastGroups, QMap<uint, uint> &oldId_To_NewId);

		bool readXMLQVApplication(QXmlStreamReader &xmlReader);
		bool readXMLNodes(QXmlStreamReader &xmlReader);
		bool readXMLNode(QXmlStreamReader &xmlReader);
		bool readXMLProperty(QXmlStreamReader &xmlReader, uint id);
		bool readXMLLinks(QXmlStreamReader &xmlReader);
		bool readXMLAsynchronousLink(QXmlStreamReader &xmlReader);
		bool readXMLSynchronousLink(QXmlStreamReader &xmlReader);
		bool readXMLSequentialLink(QXmlStreamReader &xmlReader);
		bool readXMLGroups(QXmlStreamReader &xmlReader);
		bool readXMLGroup(QXmlStreamReader &xmlReader, QList<GroupInfo> &groupInfos);
		bool readXMLChildNode(QXmlStreamReader &xmlReader, GroupInfo &groupInfo);
		bool readXMLChildGroup(QXmlStreamReader &xmlReader, GroupInfo &groupInfo);
		bool readXMLVisibleNodeLink(QXmlStreamReader &xmlReader, GroupInfo &groupInfo);
		bool readXMLVisibleGroupLink(QXmlStreamReader &xmlReader, GroupInfo &groupInfo);
		bool readXMLUnknownElement(QXmlStreamReader &xmlReader);

		ItemFactory factory;
		SlateWindow slate;
		Informer informer;

		QMap<uint, QVPropertyContainer *> containers;
		QMap<uint, QVPropertyContainer *> initialContainers;
		QMap<uint, QDialog *> dialogs; // el identificador es del container asociado
		QMap<uint, uint> dialogsToContainers; // los containers asociados a los dialogos (id dialogo, id container)

			class CreatedLink
				{
				public:
					CreatedLink(const uint _fromId, const QString _fromProp, const uint _toId, const QString _toProp, const bool _sync, const bool _sequ):
						fromId(_fromId), fromProp(_fromProp), toId(_toId), toProp(_toProp), synchronous(_sync), sequential(_sequ) { }

					uint fromId;
					QString fromProp;
					uint toId;
					QString toProp;
					bool synchronous;
					bool sequential;
				};

			class CreatedItem
				{
				public:
					CreatedItem(): type(""), name(""), id(0) { }
					CreatedItem(const QString _type, const QString _name, uint _id): type(_type), name(_name), id(_id) { }

					QString type, name;
					uint id;
				};

			class PropertyChange
				{
				public:
					PropertyChange(const uint contId, const QString prop, const QVariant val): id(contId), propName(prop), value(val) { }

					uint id;
					QString propName;
					QVariant value;
				};

		QMap<QString, CreatedLink> createdLinks; // the key is "fromName[fromProp] => toName[toProp]"
		QMap<uint, CreatedItem> createdItems;
		QMap<QString, PropertyChange> propertyChanges; // the key is "contName[propName]"

		QMap<uint, uint> XMLId_To_ContainersId; // para cuando estamos cargando el XML ya que se hace recursivo entre muchas funciones

#endif
};

#endif

